/***
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *    
 *    Linking this library statically or dynamically with other modules 
 *    is making a combined work based on this library. Thus, the terms and
 *    conditions of the GNU General Public License cover the whole
 *    combination.
 *    
 *    As a special exception, the copyright holders of this library give 
 *    you permission to link this library with independent modules to 
 *    produce an executable, regardless of the license terms of these 
 *    independent modules, and to copy and distribute the resulting 
 *    executable under terms of your choice, provided that you also meet, 
 *    for each linked independent module, the terms and conditions of the 
 *    license of that module.  An independent module is a module which 
 *    is not derived from or based on this library.  If you modify this 
 *    library, you may extend this exception to your version of the 
 *    library, but you are not obligated to do so.  If you do not wish 
 *    to do so, delete this exception statement from your version.
 *
 *    Project: www.simpledbm.org
 *    Author : Dibyendu Majumdar
 *    Email  : d dot majumdar at gmail dot com ignore
 */
package org.simpledbm.rss.impl.locking;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.LockSupport;

import org.simpledbm.common.api.exception.ExceptionHandler;
import org.simpledbm.common.api.locking.LockMode;
import org.simpledbm.common.api.platform.Platform;
import org.simpledbm.common.api.platform.PlatformObjects;
import org.simpledbm.common.api.thread.Scheduler.Priority;
import org.simpledbm.common.util.Dumpable;
import org.simpledbm.common.util.Linkable;
import org.simpledbm.common.util.SimpleLinkedList;
import org.simpledbm.common.util.SimpleTimer;
import org.simpledbm.common.util.logging.Logger;
import org.simpledbm.common.util.mcat.Message;
import org.simpledbm.common.util.mcat.MessageInstance;
import org.simpledbm.common.util.mcat.MessageType;
import org.simpledbm.rss.api.latch.Latch;
import org.simpledbm.rss.api.latch.LatchFactory;
import org.simpledbm.rss.api.locking.LockDeadlockException;
import org.simpledbm.rss.api.locking.LockDuration;
import org.simpledbm.rss.api.locking.LockException;
import org.simpledbm.rss.api.locking.LockHandle;
import org.simpledbm.rss.api.locking.LockInfo;
import org.simpledbm.rss.api.locking.LockManager;
import org.simpledbm.rss.api.locking.LockTimeoutException;

/**
 * The default implementation of the LockMgr interface is a memory based lock
 * management system modeled very closely on the description of a locking system
 * in <em>Transaction Processing: Concepts and Techniques, by Jim Gray and
 * Andreas Reuter</em>.
 * <p>
 * For each lock in the system, a queue of lock requests is maintained. The
 * queue has granted requests followed by waiting requests. To allow locks to be
 * quickly found, a hash table of all locks is maintained.
 * 
 * @author Dibyendu Majumdar
 */
public final class LockManagerImpl implements LockManager {

    /*
     * 
    Algorithm
    ---------

    The main algorithm for the lock manager is shown in the form of use cases.

    a001 - lock acquire
    *******************

    Main Success Scenario
    .....................

    1) Search for the lock header
    2) Lock header not found
    3) Allocate new lock header
    4) Allocate new lock request
    5) Append lock request to queue with status = GRANTED and reference count of 1.
    6) Set lock granted mode to GRANTED

    Extensions
    ..........

    2. a) Lock header found but client has no prior request for the lock.
      
      1. Do `a003 - handle new request`_.

    b) Lock header found and client has a prior GRANTED lock request.
      
      1. Do `a002 - handle lock conversion`_.

    a003 - handle new request
    *************************

    Main Success Scenario
    .....................

    1) Allocate new request.
    2) Append lock request to queue with reference count of 1.
    3) Check for waiting requests.
    4) Check whether request is compatible with granted mode.
    5) There are no waiting requests and lock request is compatible with
    granted mode.
    6) Set lock's granted mode to maximum of this request and existing granted mode.
    7) Success.

    Extensions
    ..........

    5. a) There are waiting requests or lock request is not compatible with
      granted mode.
      
      1. Do `a004 - lock wait`_.

    a002 - handle lock conversion
    *****************************

    Main Success Scenario
    .....................

    1) Check lock compatibility with granted group.
    2) Lock request is compatible with granted group.
    3) Grant lock request, and update granted mode for the request.

    Extensions
    ..........

    2. a) Lock request is incompatible with granted group.
      
      1. Do `a004 - lock wait`_.


    a004 - lock wait
    ****************

    Main Success Scenario
    .....................

    1) Wait for lock.
    2) Lock granted.
    3) Success.

    Extensions
    ..........

    2. a) Lock was not granted.
      
      1. Failure!

    b001 - release lock
    *******************

    Main Success Scenario
    .....................

    1) Decrease reference count.
    2) Sole lock request and reference count is zero.
    3) Remove lock header from hash table.
    4) Success.

    Extensions
    ..........

    2. a) Reference count greater than zero.
      
      1. Success.

    2. b) Reference count is zero and there are other requests on the lock.
      
      1. Remove request from the queue.
      2. Do `b002 - grant waiters`_.

    b002 - grant waiters
    ********************

    Main Success Scenario
    .....................

    1) Get next granted lock.
    2) Recalculate granted mode.
    3) Repeat from 1) until no more granted requests.
    4) Get next waiting request.
    5) Request is compatible with granted mode.
    6) Grant request and wake up thread waiting for the lock. Increment reference count of
    the granted request and set granted mode to maximum of current mode and granted request.
    7) Repeat from 4) until no more waiting requests.

    Extensions
    ..........

    1. a) Conversion request.
      
      1. Do `b003 - grant conversion request`_.
      2. Resume from 2).

    4. a) "conversion pending" is set (via b003).
      
      1. Done.

    5. a) Request is incompatible with granted mode.
      
      1. Done.


    b003 - grant conversion request
    *******************************

    Main Success Scenario
    .....................

    1) Do `c001 - check request compatible with granted group`_.
    2) Request is compatible.
    3) Grant conversion request.

    Extensions
    ..........

    2. a) Conversion request incompatible with granted group.
      
      1. Set "conversion pending" flag.

    c001 - check request compatible with granted group
    **************************************************

    Main Success Scenario
    .....................

    1) Get next granted request.
    2) Request is compatible with this request.
    3) Repeat from 1) until no more granted requests.

    Extensions
    ..........

    1. a) Request belongs to the caller.
      
      1. Resume from step 3).

    2. a) Request is incompatible with this request.
      
      1. Failure!

     */

    static final int hashPrimes[] = { 53, 97, 193, 389, 769, 1543, 3079, 6151,
            12289, 24593, 49157, 98317, 196613, 393241, 786433 };

    final Platform platform;

    final Logger log;

    final ExceptionHandler exceptionHandler;

    /**
     * Offset into {@link #hashPrimes}
     */
    private volatile int htsz = 0;

    /**
     * Tracks the number of items in the hash table
     */
    private volatile int count = 0;

    /**
     * Upper limit of number of items that can be inserted into the hash table.
     * Exceeding this causes the hash table to be resized.
     */
    private volatile int threshold = 0;

    /**
     * Used to calculate the hash table size threshold. Expressed as a
     * percentage of hash table size.
     */
    private final float loadFactor = 0.75f;

    /**
     * Hash table of locks. The hash table is dynamically resized if necessary.
     */
    private LockBucket[] LockHashTable;

    /**
     * Size of the hash table. This is always equal to hashPrimes[htsz]. Can
     * change when hash table is resized.
     */
    private int hashTableSize;

    /**
     * List of lock event listeners.
     */
    private final ArrayList<LockEventListener> lockEventListeners = new ArrayList<LockEventListener>();

    /**
     * Map of waiting lock requesters, to aid deadlock detection. Keyed by lock
     * owner.
     */
    private final Map<Object, LockWaiter> waiters = Collections
            .synchronizedMap(new HashMap<Object, LockWaiter>());

    /**
     * To keep the algorithm simple, the deadlock detector uses a global
     * exclusive lock on the lock manager. The lock manager itself acquires
     * shared locks during normal operations, thus avoiding conflict with the
     * deadlock detector.
     */
    private final Latch globalLock;

    /**
     * If set, this instructs the LockManager to stop. No further requests
     * should be accepted.
     */
    volatile boolean stop = false;

    /**
     * Handle to the deadlock detector task, which is run via the SimpleDBM
     * scheduler.
     */
    ScheduledFuture<?> deadlockDetectorThread;

    /**
     * The interval between deadlock detection scans, specified in seconds.
     * Default is 10 seconds.
     */
    int deadlockDetectorInterval = 10;

    /**
     * Cached value from logging to improve performance. Downside is that we
     * cannot change logging dynamically.
     */
    final boolean debugEnabled;

    /**
     * Cached value from logging to improve performance. Downside is that we
     * cannot change logging dynamically.
     */
    final boolean traceEnabled;

    /**
     * Defines the various lock release methods.
     * 
     * @author Dibyendu Majumdar
     */
    enum ReleaseAction {
        RELEASE, FORCE_RELEASE, DOWNGRADE;
    }

    // Lock Manager messages
    static Message m_EC0001 = new Message('R', 'C', MessageType.ERROR, 1,
            "Lock request {0} timed out");
    static Message m_WC0002 = new Message('R', 'C', MessageType.WARN, 2,
            "Lock request {0} failed due to a deadlock");
    static Message m_EC0099 = new Message('R', 'C', MessageType.ERROR, 99,
            "Unexpected error occurred while attempting to acquire lock request {0}");
    static Message m_EC0003 = new Message(
            'R',
            'C',
            MessageType.ERROR,
            3,
            "Invalid request because lock requested {0} is already being waited for by requester {1}");
    static Message m_WC0004 = new Message(
            'R',
            'C',
            MessageType.WARN,
            4,
            "Lock request {0} is not compatible with granted group {1}, timing out because this is a conditional request");
    static Message m_EC0005 = new Message('R', 'C', MessageType.ERROR, 5,
            "Unexpected error while handling conversion request");
    static Message m_EC0006 = new Message('R', 'C', MessageType.ERROR, 6,
            "Invalid call: no lock request associated with this handle");
    static Message m_EC0008 = new Message('R', 'C', MessageType.ERROR, 8,
            "Invalid request to release lock {0} as it is being waited for");
    static Message m_EC0009 = new Message('R', 'C', MessageType.ERROR, 9,
            "Invalid downgrade request: mode held {0}, mode to downgrade to {1}");
    static Message m_EC0010 = new Message('R', 'C', MessageType.ERROR, 10,
            "Listener {0} failed unexpectedly: lock parameters {1}");
    static Message m_WC0011 = new Message('R', 'C', MessageType.WARN, 11,
            "Detected deadlock cycle: {2} {3}");
    static Message m_IC0012 = new Message('R', 'C', MessageType.INFO, 12,
            "Deadlock detector STARTED");
    static Message m_IC0013 = new Message('R', 'C', MessageType.INFO, 13,
            "Deadlock detector STOPPED");
    static Message m_IC0014 = new Message('R', 'C', MessageType.INFO, 14,
            "Dumping lock table");
    static Message m_IC0015 = new Message('R', 'C', MessageType.INFO, 15,
            "LockItem = {0}");

    /**
     * Holds parameters supplied to acquire, release or find APIs.
     */
    static final class LockParameters {
        Object owner;
        Object lockable;
        LockMode mode;
        LockDuration duration;
        int timeout;
        LockInfo lockInfo;

        LockMode downgradeMode;
        ReleaseAction action;

        public String toString() {
            return "LockParameters(Lockable=" + lockable.toString()
                    + ", owner=" + owner + ", mode=" + mode + ", duration="
                    + duration + ", timeout=" + timeout + ", downgradeMode="
                    + downgradeMode + ", releaseAction=" + action + ")";
        }
    }

    /**
     * Used as a container for various lock management data.
     * 
     * @author dibyendumajumdar
     */
    static final class LockContext {
        final LockParameters parms;
        LockHandleImpl handle;
        LockRequest lockRequest;
        LockBucket bucket;
        boolean converting;
        Thread prevThread;
        LockItem lockitem;
        private LockStatus status;

        public LockContext(LockParameters parms) {
            this.parms = parms;
        }

        void setStatus(LockStatus status) {
            this.status = status;
        }

        LockStatus getStatus() {
            return status;
        }
    }

    /**
     * Creates a new LockMgrImpl, ready for use.
     */
    public LockManagerImpl(PlatformObjects po, LatchFactory latchFactory,
            Properties p) {
        platform = po.getPlatform();
        log = po.getLogger();
        exceptionHandler = po.getExceptionHandler();
        globalLock = latchFactory.newReadWriteLatch();
        htsz = 0;
        count = 0;
        hashTableSize = hashPrimes[htsz];
        LockHashTable = new LockBucket[hashTableSize];
        for (int i = 0; i < hashTableSize; i++) {
            LockHashTable[i] = getNewLockBucket();
        }
        threshold = (int) (hashTableSize * loadFactor);
        if (log.isDebugEnabled()) {
            debugEnabled = true;
        } else {
            debugEnabled = false;
        }
        if (log.isTraceEnabled()) {
            traceEnabled = true;
        } else {
            traceEnabled = false;
        }
    }

    public void start() {
        deadlockDetectorThread = platform.getScheduler()
                .scheduleWithFixedDelay(Priority.SERVER_TASK,
                        new DeadlockDetector(this), deadlockDetectorInterval,
                        deadlockDetectorInterval, TimeUnit.SECONDS);
        log.info(this.getClass().getName(), "start", new MessageInstance(
                m_IC0012).toString());
    }

    public void shutdown() {
        stop = true;
        deadlockDetectorThread.cancel(false);
        log.info(this.getClass().getName(), "shutdown", new MessageInstance(
                m_IC0013).toString());
    }

    /**
     * Grow the hash table to the next size
     */
    private void rehash() {

        if (htsz == hashPrimes.length - 1) {
            return;
        }
        if (!globalLock.tryExclusiveLock()) {
            return;
        }
        try {
            if (htsz == hashPrimes.length - 1) {
                /*
                 * Already at maximum size.
                 */
                return;
            }
            int newHashTableSize = hashPrimes[++htsz];
            if (debugEnabled) {
                log.debug(this.getClass().getName(), "rehash",
                        "SIMPLEDBM-DEBUG: Growing hash table size from "
                                + hashTableSize + " to " + newHashTableSize);
            }
            LockBucket[] newLockHashTable = new LockBucket[newHashTableSize];
            for (int i = 0; i < newHashTableSize; i++) {
                newLockHashTable[i] = getNewLockBucket();
            }
            for (int i = 0; i < hashTableSize; i++) {
                LockBucket bucket = LockHashTable[i];
                for (Iterator<LockItem> iter = bucket.chain.iterator(); iter
                        .hasNext();) {
                    LockItem item = iter.next();
                    if (item.target == null) {
                        continue;
                    }
                    int h = (item.target.hashCode() & 0x7FFFFFFF)
                            % newHashTableSize;
                    LockBucket newBucket = newLockHashTable[h];
                    newBucket.chainAppend(item);
                }
            }
            LockBucket[] oldLockHashTable = LockHashTable;
            int oldHashTableSize = hashTableSize;
            LockHashTable = newLockHashTable;
            hashTableSize = newHashTableSize;
            threshold = (int) (hashTableSize * loadFactor);
            for (int i = 0; i < oldHashTableSize; i++) {
                LockBucket bucket = oldLockHashTable[i];
                bucket.chain.clear();
                oldLockHashTable[i] = null;
            }
        } finally {
            globalLock.unlockExclusive();
        }
    }

    public void dumpLockTable() {
        globalLock.sharedLock();
        log.info(getClass().getName(), "dumpLockTable", new MessageInstance(
                m_IC0014).toString());
        try {
            for (int i = 0; i < hashTableSize; i++) {
                LockBucket bucket = LockHashTable[i];
                synchronized (bucket) {
                    for (Iterator<LockItem> iter = bucket.chain.iterator(); iter
                            .hasNext();) {
                        LockItem item = iter.next();
                        if (item.target == null) {
                            continue;
                        }
                        log.info(getClass().getName(), "dumpLockTable",
                                new MessageInstance(m_IC0015, item).toString());
                    }
                }
            }
        } finally {
            globalLock.unlockShared();
        }
    }

    public final LockMode findLock(Object owner, Object target) {

        LockParameters parms = new LockParameters();
        parms.owner = owner;
        parms.lockable = target;

        LockContext context = new LockContext(parms);

        globalLock.sharedLock();
        try {
            /* 1. Search for the lock. */
            int h = (context.parms.lockable.hashCode() & 0x7FFFFFFF)
                    % hashTableSize;
            context.lockitem = null;
            context.bucket = LockHashTable[h];
            context.lockRequest = null;
            synchronized (context.bucket) {
                context.lockitem = findLock(context);
                if (context.lockitem == null) {
                    return LockMode.NONE;
                }
                context.lockRequest = context.lockitem
                        .find(context.parms.owner);
                if (context.lockRequest != null) {
                    return context.lockRequest.mode;
                }
            }
        } finally {
            globalLock.unlockShared();
        }
        return LockMode.NONE;
    }

    /**
     * Search for the specified lockable object. Garbage collect any items that
     * are no longer needed.
     */
    private LockItem findLock(LockContext lockState) {
        for (Iterator<LockItem> iter = lockState.bucket.chain.iterator(); iter
                .hasNext();) {
            LockItem item = iter.next();
            if (item.target == null) {
                iter.remove();
                continue;
            }
            if (lockState.parms.lockable == item.target
                    || lockState.parms.lockable.equals(item.target)) {
                return item;
            }
        }
        return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.simpledbm.locking.LockMgr#acquire(java.lang.Object,
     *      java.lang.Object, org.simpledbm.locking.LockMode,
     *      org.simpledbm.locking.LockDuration, int)
     */
    public final LockHandle acquire(Object owner, Object target, LockMode mode,
            LockDuration duration, int timeout, LockInfo lockInfo) {

        LockParameters parms = new LockParameters();
        parms.owner = owner;
        parms.lockable = target;
        parms.mode = mode;
        parms.duration = duration;
        parms.timeout = timeout;
        parms.lockInfo = lockInfo;

        LockContext context = new LockContext(parms);

        if (count > threshold) {
            rehash();
        }

        globalLock.sharedLock();
        LockHandleImpl handle = null;
        try {
            handle = doAcquire(context);
            if (duration == LockDuration.INSTANT_DURATION
                    && context.getStatus() == LockStatus.GRANTED) {
                /*
                 * Handle the case where the lock was granted after a wait.
                 */
                handle.release(false);
            }
        } finally {
            globalLock.unlockShared();
        }
        return handle;
    }

    /**
     * Checks whether the specified lock request is compatible with the granted
     * group. Also sets the otherHolders flag if the granted group contains
     * other requests.
     */
    private boolean checkCompatible(LockItem lock, LockRequest request,
            LockMode mode, LockInfo lockInfo) {

        if (lockInfo != null) {
            lockInfo.setHeldByOthers(false);
        }
        boolean iscompatible = true;

        /* Check if there are other holders */
        for (LockRequest other : lock.getQueue()) {

            if (other == request)
                continue;
            else if (other.status == LockRequestStatus.WAITING)
                break;
            else {
                if (lockInfo != null) {
                    lockInfo.setHeldByOthers(true);
                }
                if (!mode.isCompatible(other.mode)) {
                    iscompatible = false;
                    break;
                }
            }
        }
        return iscompatible;
    }

    /**
     * Acquires a lock in the specified mode. Handles most of the cases except
     * the case where an INSTANT_DURATION lock needs to be waited for. This case
     * requires the lock to be released after it has been granted; the lock
     * release is handled by {@link #acquire acquire}.
     */
    private LockHandleImpl doAcquire(LockContext context) {

        if (debugEnabled) {
            log.debug(getClass().getName(), "acquire",
                    "SIMPLEDBM-DEBUG: Lock requested by " + context.parms.owner
                            + " for " + context.parms.lockable + ", mode="
                            + context.parms.mode + ", duration="
                            + context.parms.duration);
        }

        context.handle = new LockHandleImpl(this, context.parms);
        context.converting = false;
        context.prevThread = Thread.currentThread();

        /* 1. Search for the lock. */
        int h = (context.parms.lockable.hashCode() & 0x7FFFFFFF)
                % hashTableSize;
        context.lockitem = null;
        context.bucket = LockHashTable[h];
        context.lockRequest = null;
        synchronized (context.bucket) {

            context.lockitem = findLock(context);
            if (context.lockitem == null) {
                /*
                 * 2. If not found, this is a new lock and therefore grant the
                 * lock, and return success.
                 */
                handleNewLock(context);
                return context.handle;
            }

            /*
             * 3. Else check if requesting transaction already has a lock
             * request.
             */
            context.lockRequest = context.lockitem.find(context.parms.owner);

            if (context.lockRequest == null) {
                if (handleNewRequest(context)) {
                    return context.handle;
                }
            } else {
                context.handle.lockRequest = context.lockRequest;
                if (handleConversionRequest(context)) {
                    return context.handle;
                }
            }

            /* 8. Wait for the lock to be available/compatible. */
            prepareToWait(context);
        }
        notifyLockEventListeners(context);
        /*
         * Add request to list of waiters to allow deadlock detection.
         */
        LockWaiter waiter = new LockWaiter(context.lockRequest, Thread
                .currentThread());
        waiters.put(context.lockRequest.owner, waiter);
        /*
         * Global lock is released only after the waiter list has been updated.
         */
        SimpleTimer timer = new SimpleTimer((context.parms.timeout < 0) ? -1
                : TimeUnit.NANOSECONDS.convert(context.parms.timeout,
                        TimeUnit.SECONDS));
        for (;;) {
            globalLock.unlockShared();
            try {
                timer.await();
            } finally {
                globalLock.sharedLock();
            }
            /*
             * As the hash table may have been resized while we were waiting, we
             * need to recalculate the bucket.
             */
            h = (context.parms.lockable.hashCode() & 0x7FFFFFFF)
                    % hashTableSize;
            context.bucket = LockHashTable[h];
            synchronized (context.bucket) {
                if (context.lockRequest.status == LockRequestStatus.WAITING
                        || context.lockRequest.status == LockRequestStatus.CONVERTING) {
                    if (!timer.isExpired()) {
                        continue;
                    }
                }
                waiters.remove(context.lockRequest.owner);
                handleWaitResult(context);
                return context.handle;
            }
        }
    }

    /**
     * Handles the case where there aren't any locks on the target.
     */
    private void handleNewLock(LockContext context) {
        if (debugEnabled) {
            log
                    .debug(getClass().getName(), "acquire",
                            "SIMPLEDBM-DEBUG: Lock not found, therefore granting immediately");
        }
        if (context.parms.duration != LockDuration.INSTANT_DURATION) {
            LockItem lockitem = getNewLockItem(context.parms.lockable,
                    context.parms.mode);
            LockRequest r = new LockRequest(lockitem, context.parms.owner,
                    context.parms.mode, context.parms.duration);
            lockitem.queueAppend(r);
            context.bucket.chainAppend(lockitem);
            count++;
            context.handle.lockRequest = r;
            context.setStatus(LockStatus.GRANTED);
        } else {
            context.setStatus(LockStatus.GRANTABLE);
        }
    }

    private LockBucket getNewLockBucket() {
        return new LockBucket();
    }

    private LockItem getNewLockItem(Object target, LockMode mode) {
        return new LockItem(target, mode);
    }

    /**
     * Handles a new lock request when the lock is already held by some other.
     * 
     * @return true if the lock request was processed, else false to indicate
     *         that the requester must wait
     */
    private boolean handleNewRequest(LockContext context) {
        /* 4. If not, this is the first request by the transaction. */
        if (debugEnabled) {
            log.debug(getClass().getName(), "handleNewRequest",
                    "SIMPLEDBM-DEBUG: New request by transaction "
                            + context.parms.owner + " for target "
                            + context.parms.lockable);
        }

        if (context.parms.lockInfo != null) {
            context.parms.lockInfo.setHeldByOthers(true);
        }
        /*
         * 5. Check if lock can be granted. This is true if there are no
         * waiting requests and the new request is compatible with
         * existing grant mode.
         */
        boolean can_grant = (!context.lockitem.waiting && context.parms.mode
                .isCompatible(context.lockitem.grantedMode));

        if (context.parms.duration == LockDuration.INSTANT_DURATION
                && can_grant) {
            /* 6. If yes, grant the lock and return success. */
            context.setStatus(LockStatus.GRANTABLE);
            return true;
        }

        else if (!can_grant && context.parms.timeout == 0) {
            /* 7. Otherwise, if nowait was specified, return failure. */
            throw new LockTimeoutException(new MessageInstance(m_WC0004,
                    context.parms, context.lockitem));
        }

        /* Allocate new lock request */
        context.lockRequest = new LockRequest(context.lockitem,
                context.parms.owner, context.parms.mode, context.parms.duration);
        context.lockitem.queueAppend(context.lockRequest);
        context.handle.lockRequest = context.lockRequest;
        if (can_grant) {
            /* 6. If yes, grant the lock and return success. */
            if (debugEnabled) {
                log.debug(getClass().getName(), "handleNewRequest",
                        "SIMPLEDBM-DEBUG: There are no waiting locks and request is compatible with  "
                                + context.lockitem
                                + ", therefore granting lock");
            }
            context.lockitem.grantedMode = context.parms.mode
                    .maximumOf(context.lockitem.grantedMode);
            context.setStatus(LockStatus.GRANTED);
            return true;
        } else {
            context.converting = false;
            return false;
        }
    }

    /**
     * Handles a conversion request in the nowait situation.
     * 
     * @return true if conversion request was handled else false to indicate
     *         that requester must enter wait.
     */
    private boolean handleConversionRequest(LockContext context) {
        /*
         * 11. If calling transaction already has a granted lock request
         * then this must be a conversion request.
         */
        if (traceEnabled) {
            log.trace(getClass().getName(), "handleConversionRequest",
                    "SIMPLEDBM-DEBUG: Lock conversion request by transaction "
                            + context.parms.owner + " for target "
                            + context.parms.lockable);
        }

        /*
         * Limitation: a transaction cannot attempt to lock an object
         * for which it is already waiting.
         */
        if (context.lockRequest.status == LockRequestStatus.CONVERTING
                || context.lockRequest.status == LockRequestStatus.WAITING) {
            exceptionHandler.errorThrow(this.getClass().getName(),
                    "handleConversionRequest",
                    new LockException(new MessageInstance(m_EC0003,
                            context.parms.lockable, context.parms.owner)));
        }

        else if (context.lockRequest.status == LockRequestStatus.GRANTED) {
            /*
             * 12. Check whether the new request lock is same mode as
             * previously held lock.
             */
            if (context.parms.lockInfo != null) {
                context.parms.lockInfo
                        .setPreviousMode(context.lockRequest.mode);
            }
            if (context.parms.mode == context.lockRequest.mode) {
                /* 13. If so, grant lock and return. */
                if (debugEnabled) {
                    log
                            .debug(
                                    getClass().getName(),
                                    "handleConversionRequest",
                                    "SIMPLEDBM-DEBUG: Requested mode is the same as currently held mode, therefore granting");
                }
                if (context.parms.duration != LockDuration.INSTANT_DURATION) {
                    context.lockRequest.count++;
                }
                checkCompatible(context.lockitem, context.lockRequest,
                        context.parms.mode, context.parms.lockInfo);
                if (context.parms.duration == LockDuration.INSTANT_DURATION) {
                    context.setStatus(LockStatus.GRANTABLE);
                } else {
                    if (context.lockRequest.duration == LockDuration.MANUAL_DURATION
                            && context.parms.duration == LockDuration.COMMIT_DURATION) {
                        context.lockRequest.duration = LockDuration.COMMIT_DURATION;
                    }
                    context.setStatus(LockStatus.GRANTED);
                }
                return true;
            }

            else {
                /*
                 * 14. Otherwise, check if requested lock is compatible
                 * with granted group
                 */
                boolean can_grant = checkCompatible(context.lockitem,
                        context.lockRequest, context.parms.mode,
                        context.parms.lockInfo);

                if (can_grant) {
                    /* 13. If so, grant lock and return. */
                    if (debugEnabled) {
                        log.debug(getClass().getName(),
                                "handleConversionRequest",
                                "SIMPLEDBM-DEBUG: Conversion request is compatible with granted group "
                                        + context.lockitem
                                        + ", therefore granting");
                    }
                    if (context.parms.duration != LockDuration.INSTANT_DURATION) {
                        context.lockRequest.mode = context.parms.mode
                                .maximumOf(context.lockRequest.mode);
                        context.lockRequest.count++;
                        if (context.lockRequest.duration == LockDuration.MANUAL_DURATION
                                && context.parms.duration == LockDuration.COMMIT_DURATION) {
                            context.lockRequest.duration = LockDuration.COMMIT_DURATION;
                        }
                        context.lockitem.grantedMode = context.lockRequest.mode
                                .maximumOf(context.lockitem.grantedMode);
                        context.setStatus(LockStatus.GRANTED);
                    } else {
                        context.setStatus(LockStatus.GRANTABLE);
                    }
                    return true;
                }

                else if (!can_grant && context.parms.timeout == 0) {
                    /* 15. If not, and nowait specified, return failure. */
                    throw new LockTimeoutException(new MessageInstance(
                            m_WC0004, context.parms, context.lockitem));
                }

                else {
                    context.converting = true;
                    return false;
                }
            }
        } else {
            exceptionHandler.errorThrow(this.getClass().getName(),
                    "handleConversionRequest", new LockException(
                            new MessageInstance(m_EC0005)));
        }
        return false;
    }

    /**
     * Prepare the lock request for waiting.
     */
    private void prepareToWait(LockContext context) {
        context.lockitem.waiting = true;
        if (!context.converting) {
            if (debugEnabled) {
                log.debug(getClass().getName(), "prepareToWait",
                        "SIMPLEDBM-DEBUG: Waiting for lock to be free");
            }
            context.lockRequest.status = LockRequestStatus.WAITING;
        } else {
            if (debugEnabled) {
                log
                        .debug(
                                getClass().getName(),
                                "prepareToWait",
                                "SIMPLEDBM-DEBUG: Conversion NOT compatible with granted group, therefore waiting ...");
            }
            context.lockRequest.convertMode = context.parms.mode;
            context.lockRequest.convertDuration = context.parms.duration;
            context.lockRequest.status = LockRequestStatus.CONVERTING;
            context.prevThread = context.lockRequest.thread;
            context.lockRequest.thread = Thread.currentThread();
        }
    }

    /**
     * Handles the result of a lock wait.
     */
    private void handleWaitResult(LockContext context) {
        LockRequestStatus lockRequestStatus = context.lockRequest.status;
        if (lockRequestStatus == LockRequestStatus.GRANTED) {
            context.setStatus(LockStatus.GRANTED);
        } else if (lockRequestStatus == LockRequestStatus.DENIED) {
            context.setStatus(LockStatus.DEADLOCK);
        } else {
            context.setStatus(LockStatus.TIMEOUT);
        }

        if (context.getStatus() == LockStatus.GRANTED) {
            /*
             * 9. If after the wait, the lock has been granted, then return
             * success.
             */
            if (debugEnabled) {
                log.debug(getClass().getName(), "handleWaitResult",
                        "SIMPLEDBM-DEBUG: Woken up, and lock granted");
            }
            checkCompatible(context.lockitem, context.lockRequest,
                    context.parms.mode, context.parms.lockInfo);
            return;
        }

        /* 10. Else return failure. */
        if (debugEnabled) {
            log.debug(getClass().getName(), "handleWaitResult",
                    "SIMPLEDBM-DEBUG: Woken up, and lock failed");
        }

        if (!context.converting) {
            /* If not converting the delete the newly created request. */
            context.lockitem.queueRemove(context.lockRequest);
            if (context.lockitem.queueIsEmpty()) {
                context.lockitem.reset(); // Setup lock for garbage collection
                count--;
            }
        } else {
            /* If converting, then restore old status */
            context.lockRequest.status = LockRequestStatus.GRANTED;
            context.lockRequest.convertMode = context.lockRequest.mode;
            context.lockRequest.thread = context.prevThread;
        }
        if (context.getStatus() == LockStatus.DEADLOCK) {
            /* 
             * If we have been chosen as a deadlock victim, then we need to grant the
             * lock to the waiter who has won the deadlock.
             */
            grantWaiters(ReleaseAction.RELEASE, context.lockRequest,
                    context.handle, context.lockitem, context.parms.lockInfo);
        }
        if (context.getStatus() == LockStatus.TIMEOUT) {
            exceptionHandler.warnAndThrow(this.getClass().getName(),
                    "handleWaitResult", new LockTimeoutException(
                            new MessageInstance(m_EC0001, context.parms)));
        } else if (context.getStatus() == LockStatus.DEADLOCK) {
            throw new LockDeadlockException(new MessageInstance(m_WC0002,
                    context.parms));
        } else {
            exceptionHandler.warnAndThrow(this.getClass().getName(),
                    "handleWaitResult", new LockException(new MessageInstance(
                            m_EC0099, context.parms)));
        }
    }

    public boolean release(Object owner, Object lockable, boolean force) {
        LockParameters parms = new LockParameters();
        parms.lockable = lockable;
        parms.owner = owner;
        parms.action = force ? ReleaseAction.FORCE_RELEASE
                : ReleaseAction.RELEASE;

        LockContext context = new LockContext(parms);
        context.handle = new LockHandleImpl(this, context.parms);

        globalLock.sharedLock();
        try {
            return doReleaseInternal(context);
        } finally {
            globalLock.unlockShared();
        }
    }

    public boolean downgrade(Object owner, Object lockable, LockMode downgradeTo) {
        LockParameters parms = new LockParameters();
        parms.lockable = lockable;
        parms.owner = owner;
        parms.action = ReleaseAction.DOWNGRADE;
        parms.downgradeMode = downgradeTo;

        LockContext context = new LockContext(parms);
        context.handle = new LockHandleImpl(this, context.parms);

        globalLock.sharedLock();
        try {
            return doReleaseInternal(context);
        } finally {
            globalLock.unlockShared();
        }
    }

    /**
     * Release or downgrade a specified lock.
     */
    private boolean doRelease(LockHandle handle, ReleaseAction action,
            LockMode downgradeMode) {

        LockParameters parms = new LockParameters();
        LockHandleImpl handleImpl = (LockHandleImpl) handle;
        parms.lockable = handleImpl.getLockable();
        parms.owner = handleImpl.getOwner();
        parms.action = action;
        parms.downgradeMode = downgradeMode;

        LockContext context = new LockContext(parms);
        context.handle = handleImpl;

        globalLock.sharedLock();
        try {
            return doReleaseInternal(context);
        } finally {
            globalLock.unlockShared();
        }
    }

    private boolean doReleaseInternal(LockContext context) {
        context.lockRequest = null;
        boolean released = false;

        if (debugEnabled) {
            log.debug(getClass().getName(), "doReleaseInternal",
                    "SIMPLEDBM-DEBUG: Request by " + context.parms.owner
                            + " to release lock for " + context.parms.lockable);
        }
        int h = (context.parms.lockable.hashCode() & 0x7FFFFFFF)
                % hashTableSize;
        context.bucket = LockHashTable[h];
        synchronized (context.bucket) {
            /* 1. Search for the lock. */
            context.lockitem = findLock(context);

            if (context.lockitem == null) {
                /* 2. If not found, return success. */
                /*
                 * Rather than throwing an exception, we return success. This allows
                 * us to use this method in situations where for reasons of efficiency,
                 * the client cannot track the status of lock objects, and therefore
                 * may end up trying to release the same lock multiple times.
                 */
                return true;
            }
            released = releaseLock(context);
        }
        return released;
    }

    /**
     * Release or downgrade a specified lock.
     * 
     * <p>
     * Algorithm:
     * <ol>
     * <li>1. Search for the lock.</li>
     * <li>2. If not found, return Ok.</li>
     * <li>3. If found, look for the transaction's lock request.</li>
     * <li>4. If not found, return Ok.</li>
     * <li>5. If lock request is in invalid state, return error.</li>
     * <li>6. If noforce and not downgrading, and reference count greater than
     * 0, then do not release the lock request. Decrement reference count and
     * return Ok.</li>
     * <li>7. If sole lock request and not downgrading, then release the lock
     * and return Ok.</li>
     * <li>8. If not downgrading, delete the lock request from the queue.
     * Otherwise, downgrade the mode assigned to the lock request.
     * 
     * </li>
     * <li>9. Recalculate granted mode by calculating max mode amongst all
     * granted (including conversion) requests. If a conversion request is
     * compatible with all other granted requests, then grant the conversion,
     * recalculating granted mode. If a waiting request is compatible with
     * granted mode, and there are no pending conversion requests, then grant
     * the request, and recalculate granted mode. Otherwise, we are done.</li>
     * </ol>
     * </p>
     * <p>
     * Note that this means that FIFO is respected for waiting requests, but
     * conversion requests are granted as soon as they become compatible. Also,
     * note that if a conversion request is pending, waiting requests cannot be
     * granted.
     * </p>
     * </p>
     */
    private boolean releaseLock(LockContext context) {
        boolean released;
        /* 3. If lock found, look for the transaction's lock request. */
        context.lockRequest = context.lockitem.find(context.parms.owner);

        if (context.lockRequest == null) {
            /* 4. If not found, return success. */
            /*
             * Rather than throwing an exception, we return success. This allows
             * us to use this method in situations where for reasons of efficiency,
             * the client cannot track the status of lock objects, and therefore
             * may end up trying to release the same lock multiple times.
             */
            if (debugEnabled) {
                log
                        .debug(getClass().getName(), "release",
                                "SIMPLEDBM-DEBUG: request not found, returning success");
            }
            return true;
        }

        if (context.lockRequest.status == LockRequestStatus.CONVERTING
                || context.lockRequest.status == LockRequestStatus.WAITING) {
            /* 5. If lock in invalid state, return error. */
            exceptionHandler.errorThrow(this.getClass().getName(),
                    "releaseLock", new LockException(new MessageInstance(
                            m_EC0008, context.lockRequest)));
        }

        if (context.parms.action == ReleaseAction.DOWNGRADE
                && context.lockRequest.mode == context.parms.downgradeMode) {
            /*
             * If downgrade request and lock is already in target mode,
             * return success.
             */
            return true;
        }

        if (context.parms.action == ReleaseAction.RELEASE
                && context.lockRequest.duration == LockDuration.COMMIT_DURATION) {
            /*
             * 6(1). If noforce, and lock is held for commit duration, then do
             * not release the lock request. 
             */
            if (debugEnabled) {
                log
                        .debug(getClass().getName(), "release",
                                "SIMPLEDBM-DEBUG: Lock not released as it is held for commit duration");
            }
            return false;
        }

        if (context.parms.action == ReleaseAction.RELEASE
                && context.lockRequest.count > 1) {
            /*
             * 6. If noforce, and reference count greater than 0, then do
             * not release the lock request. Decrement reference count if
             * greater than 0, and, return Ok.
             */
            if (debugEnabled) {
                log
                        .debug(getClass().getName(), "release",
                                "SIMPLEDBM-DEBUG: Count decremented but lock not released");
            }
            context.lockRequest.count--;
            return false;
        }

        /*
         * Either the lock is being downgraded or it is being released and
         * its reference count == 0 or it is being forcibly released.
         */

        if (context.lockRequest == context.lockitem.queueHead()
                && context.lockRequest == context.lockitem.queueTail()
                && context.parms.action != ReleaseAction.DOWNGRADE) {
            /* 7. If sole lock request, then release the lock and return Ok. */
            if (debugEnabled) {
                log
                        .debug(getClass().getName(), "release",
                                "SIMPLEDBM-DEBUG: Removing sole lock, releasing lock object");
            }
            context.lockitem.queueRemove(context.lockRequest);
            context.lockitem.reset();
            count--;
            return true;
        }

        /*
         * 8. If not downgrading, delete the lock request from the queue.
         * Otherwise, downgrade the mode assigned to the lock request.
         */
        if (context.parms.action != ReleaseAction.DOWNGRADE) {
            if (debugEnabled) {
                log.debug(getClass().getName(), "release",
                        "SIMPLEDBM-DEBUG: Removing lock request "
                                + context.lockRequest
                                + " and re-adjusting granted mode");
            }
            context.lockitem.queueRemove(context.lockRequest);
            released = true;
        } else {
            /*
             * We need to determine whether is a valid downgrade request. To
             * do so, we do a reverse check - ie, if the new mode could have
             * been upgraded to current mode, then it is okay to downgrade.
             */
            LockMode mode = context.parms.downgradeMode
                    .maximumOf(context.lockRequest.mode);
            if (mode == context.lockRequest.mode) {
                if (debugEnabled) {
                    log.debug(getClass().getName(), "release",
                            "SIMPLEDBM-DEBUG: Downgrading "
                                    + context.lockRequest + " to "
                                    + context.parms.downgradeMode
                                    + " and re-adjusting granted mode");
                }
                // FIXME: We do not need to set LockInfo here anymore 
                if (context.parms.lockInfo != null) {
                    context.parms.lockInfo
                            .setPreviousMode(context.lockRequest.mode);
                    context.parms.lockInfo.setHeldByOthers(false);
                }
                context.lockRequest.convertMode = context.lockRequest.mode = context.parms.downgradeMode;
            } else {
                exceptionHandler.errorThrow(this.getClass().getName(),
                        "releaseLock", new LockException(new MessageInstance(
                                m_EC0009, context.lockRequest.mode,
                                context.parms.downgradeMode)));
            }
            released = false;
        }
        /*
         * 9. Recalculate granted mode by calculating max mode amongst all
         * granted (including conversion) requests. 
         */
        grantWaiters(context.parms.action, context.lockRequest, context.handle,
                context.lockitem, context.parms.lockInfo);
        return released;
    }

    private void grantWaiters(ReleaseAction action, LockRequest myReq,
            LockHandleImpl handleImpl, LockItem lockitem, LockInfo lockInfo) {
        /*
         * 9. Recalculate granted mode by calculating max mode amongst all
         * granted (including conversion) requests. If a conversion request
         * is compatible with all other granted requests, then grant the
         * conversion, recalculating granted mode. If a waiting request is
         * compatible with granted mode, and there are no pending conversion
         * requests, then grant the request, and recalculate granted mode.
         * Otherwise, we are done. Note that this means that FIFO is
         * respected for waiting requests, but conversion requests are
         * granted as soon as they become compatible. Also, note that if a
         * conversion request is pending, waiting requests cannot be
         * granted.
         */
        boolean converting;
        lockitem.grantedMode = LockMode.NONE;
        lockitem.waiting = false;
        converting = false;
        for (LockRequest r : lockitem.getQueue()) {

            if (r.status == LockRequestStatus.GRANTED) {
                lockitem.grantedMode = r.mode.maximumOf(lockitem.grantedMode);
                if (r != myReq && action == ReleaseAction.DOWNGRADE) {
                    if (lockInfo != null) {
                        lockInfo.setHeldByOthers(true);
                    }
                }
            }

            else if (r.status == LockRequestStatus.CONVERTING) {
                boolean can_grant;

                assert (!converting || lockitem.waiting);

                can_grant = checkCompatible(lockitem, r, r.convertMode, null);
                if (can_grant) {
                    if (debugEnabled) {
                        log
                                .debug(
                                        getClass().getName(),
                                        "release",
                                        "SIMPLEDBM-DEBUG: Granting conversion request "
                                                + r
                                                + " because request is compatible with "
                                                + lockitem);
                    }
                    if (r.convertDuration == LockDuration.INSTANT_DURATION) {
                        /*
                         * If the request is for an instant duration lock then
                         * don't perform the conversion.
                         */
                        r.convertMode = r.mode;
                    } else {
                        r.mode = r.convertMode.maximumOf(r.mode);
                        r.convertMode = r.mode;
                        if (r.convertDuration == LockDuration.COMMIT_DURATION
                                && r.duration == LockDuration.MANUAL_DURATION) {
                            r.duration = LockDuration.COMMIT_DURATION;
                        }
                        lockitem.grantedMode = r.mode
                                .maximumOf(lockitem.grantedMode);
                    }
                    /*
                     * Treat conversions as lock recursion.
                     */
                    r.count++;
                    r.status = LockRequestStatus.GRANTED;
                    LockSupport.unpark(r.thread);
                } else {
                    lockitem.grantedMode = r.mode
                            .maximumOf(lockitem.grantedMode);
                    converting = true;
                    lockitem.waiting = true;
                }
            }

            else if (r.status == LockRequestStatus.WAITING) {
                if (!converting && r.mode.isCompatible(lockitem.grantedMode)) {
                    if (debugEnabled) {
                        log
                                .debug(
                                        getClass().getName(),
                                        "release",
                                        "SIMPLEDBM-DEBUG: Granting waiting request "
                                                + r
                                                + " because not converting and request is compatible with "
                                                + lockitem);
                    }
                    r.status = LockRequestStatus.GRANTED;
                    lockitem.grantedMode = r.mode
                            .maximumOf(lockitem.grantedMode);
                    LockSupport.unpark(r.thread);
                } else {
                    if (debugEnabled && converting) {
                        log
                                .debug(
                                        getClass().getName(),
                                        "release",
                                        "SIMPLEDBM-DEBUG: Cannot grant waiting request "
                                                + r
                                                + " because conversion request pending");
                    }
                    lockitem.waiting = true;
                    break;
                }
            }
        }
    }

    /* (non-Javadoc)
     * @see org.simpledbm.rss.api.locking.LockManager#getLocks(java.lang.Object, org.simpledbm.rss.api.locking.LockMode)
     */
    public Object[] getLocks(Object owner, LockMode mode) {
        ArrayList<Object> locks = new ArrayList<Object>();
        globalLock.sharedLock();
        try {
            for (int i = 0; i < LockHashTable.length; i++) {
                LockBucket bucket = LockHashTable[i];
                synchronized (bucket) {
                    for (Iterator<LockItem> iter = bucket.chain.iterator(); iter
                            .hasNext();) {
                        LockItem item = iter.next();
                        if (item.target == null) {
                            continue;
                        }
                        LockRequest lockRequest = item.find(owner);
                        if (lockRequest != null
                                && lockRequest.status == LockRequestStatus.GRANTED) {
                            if (mode == null || lockRequest.mode == mode) {
                                locks.add(item.target);
                            }
                        }
                    }
                }
            }
            return locks.toArray();
        } finally {
            globalLock.unlockShared();
        }
    }

    public synchronized void addLockEventListener(LockEventListener listener) {
        lockEventListeners.add(listener);
    }

    public synchronized void clearLockEventListeners() {
        lockEventListeners.clear();
    }

    private void notifyLockEventListeners(LockContext context) {
        for (LockEventListener listener : lockEventListeners) {
            try {
                listener.beforeLockWait(context.parms.owner,
                        context.parms.lockable, context.parms.mode);
            } catch (Exception e) {
                log.error(this.getClass().getName(),
                        "notifyLockEventListeners", new MessageInstance(
                                m_EC0010, listener, context.parms).toString(),
                        e);
            }
        }
    }

    private boolean findDeadlockCycle(LockWaiter me) {
        if (me.visited) {
            return false;
        } else {
            me.visited = true;
        }
        LockWaiter him;

        assert me.cycle == null;
        assert me.myLockRequest.status == LockRequestStatus.WAITING
                || me.myLockRequest.status == LockRequestStatus.CONVERTING;

        if (me.myLockRequest.status == LockRequestStatus.CONVERTING) {
            LockMode mode = me.myLockRequest.convertMode;

            /*
             * Look at everyone that holds the lock
             */
            for (LockRequest them : me.myLockRequest.lockItem.queue) {
                if (them.owner == me.myLockRequest.owner) {
                    continue;
                }
                if (them.status == LockRequestStatus.DENIED) {
                    continue;
                }
                if (them.status == LockRequestStatus.WAITING) {
                    break;
                }
                boolean incompatible = !them.mode.isCompatible(mode);
                if (incompatible) {
                    him = waiters.get(them.owner);
                    if (him == null) {
                        return false;
                    }
                    me.cycle = him;
                    if (him.cycle != null) {
                        /*
                         * Choose the victim:
                         * Prefer the manual duration locker to be a victim
                         */
                        if (him.myLockRequest.duration == LockDuration.MANUAL_DURATION
                                && me.myLockRequest.duration == LockDuration.COMMIT_DURATION) {
                            /*
                             * Swap him and me.
                             */
                            LockWaiter temp = me;
                            me = him;
                            him = temp;
                        }
                        log.warn(log.getClass().getName(), "findDeadlockCycle",
                                new MessageInstance(m_WC0011, me.myLockRequest,
                                        him.myLockRequest,
                                        me.myLockRequest.lockItem,
                                        him.myLockRequest.lockItem).toString());

                        me.myLockRequest.status = LockRequestStatus.DENIED;
                        LockSupport.unpark(me.thread);
                        return true;
                    }
                    boolean result = findDeadlockCycle(him);
                    //me.cycle = null;
                    if (result) {
                        return result;
                    }
                }
            }
        } else if (me.myLockRequest.status == LockRequestStatus.WAITING) {
            LockMode mode = me.myLockRequest.mode;

            /*
             * Look at everyone ahead of me in the queue
             */
            for (LockRequest them : me.myLockRequest.lockItem.queue) {
                if (them.owner == me.myLockRequest.owner) {
                    break;
                }
                if (them.status == LockRequestStatus.DENIED) {
                    continue;
                }
                boolean incompatible = !them.mode.isCompatible(mode);
                if (incompatible || them.status == LockRequestStatus.CONVERTING
                        || them.status == LockRequestStatus.WAITING) {
                    him = waiters.get(them.owner);
                    if (him == null) {
                        return false;
                    }
                    me.cycle = him;
                    if (him.cycle != null) {
                        /*
                         * Choose the victim:
                         * Prefer the manual duration locker to be a victim
                         */
                        if (him.myLockRequest.duration == LockDuration.MANUAL_DURATION
                                && me.myLockRequest.duration == LockDuration.COMMIT_DURATION) {
                            /*
                             * Swap him and me.
                             */
                            LockWaiter temp = me;
                            me = him;
                            him = temp;
                        }
                        log.warn(log.getClass().getName(), "findDeadlockCycle",
                                new MessageInstance(m_WC0011, me.myLockRequest,
                                        him.myLockRequest,
                                        me.myLockRequest.lockItem,
                                        him.myLockRequest.lockItem).toString());
                        me.myLockRequest.status = LockRequestStatus.DENIED;
                        LockSupport.unpark(me.thread);
                        return true;
                    }
                    boolean result = findDeadlockCycle(him);
                    //me.cycle = null;
                    if (result) {
                        return result;
                    }
                }
            }
        }

        me.cycle = null;
        return false;
    }

    void detectDeadlocks() {
        /*
         * The deadlock detector is a very simple implementation
         * based upon example shown in the Transaction Processing,
         * by Jim Gray and Andreas Reuter.
         * See sections 7.11.3 and section 8.5.
         */
        int retry = 0;
        for (; retry < 5; retry++) {
            if (globalLock.tryExclusiveLock()) {
                break;
            }
            Thread.yield();
        }
        if (retry == 5) {
            return;
        }
        try {
            LockWaiter[] waiterArray = waiters.values().toArray(
                    new LockWaiter[0]);
            for (LockWaiter waiter : waiterArray) {
                waiter.cycle = null;
                waiter.visited = false;
            }
            for (LockWaiter waiter : waiterArray) {
                findDeadlockCycle(waiter);
            }
        } finally {
            globalLock.unlockExclusive();
        }
    }

    static class LockBucket {

//        final LinkedList<LockItem> chain = new LinkedList<LockItem>();
        final SimpleLinkedList<LockItem> chain = new SimpleLinkedList<LockItem>();

        LockItem chainHead() {
            return chain.getFirst();
        }

        LockItem chainTail() {
            return chain.getLast();
        }

        void chainAppend(LockItem item) {
//            chain.add(item);
            chain.addLast(item);
        }

        void chainRemove(LockItem item) {
            chain.remove(item);
        }

    }

    /**
     * LockItem is the lock header for each lockable target. Against each
     * lockable object, there is a queue of lock requests, a granted mode, and
     * waiting status.
     * 
     * @author Dibyendu
     */
    static class LockItem extends Linkable implements Dumpable {

        Object target;

//        final LinkedList<LockRequest> queue = new LinkedList<LockRequest>();

        final SimpleLinkedList<LockRequest> queue = new SimpleLinkedList<LockRequest>();
        
        
        LockMode grantedMode;

        boolean waiting;

        LockItem(Object target, LockMode mode) {
            this.target = target;
            this.grantedMode = mode;
        }

        void setLockMode(LockMode mode) {
            this.grantedMode = mode;
        }

        LockMode getLockMode() {
            return grantedMode;
        }

        Object getTarget() {
            return target;
        }

        boolean isWaiting() {
            return waiting;
        }

        void setWaiting(boolean waiting) {
            this.waiting = waiting;
        }

        void queueAppend(LockRequest request) {
//            queue.add(request);
            queue.addLast(request);
        }

        void queueRemove(LockRequest request) {
            queue.remove(request);
        }

        LockRequest queueHead() {
            return queue.getFirst();
        }

        LockRequest queueTail() {
            return queue.getLast();
        }

        boolean queueIsEmpty() {
            return queue.isEmpty();
        }

//        LinkedList<LockRequest> getQueue() {
        SimpleLinkedList<LockRequest> getQueue() {
            return queue;
        }

        LockRequest find(Object owner) {
            for (LockRequest req : queue) {
                if (req.owner == owner || req.owner.equals(owner)) {
                    return req;
                }
            }
            return null;
        }

        void reset() {
            target = null;
            grantedMode = LockMode.NONE;
        }

        @Override
        public String toString() {
            StringBuilder sb = new StringBuilder();
            return appendTo(sb).toString();
        }

        public StringBuilder appendTo(StringBuilder sb) {
            sb.append(newline).append("LockItem(").append(newline).append(TAB)
                    .append("target=");
            if (target instanceof Dumpable) {
                ((Dumpable) target).appendTo(sb);
            } else {
                sb.append(target);
            }
            sb.append(newline).append(TAB).append("granted mode=").append(
                    grantedMode).append(newline);
            sb.append(TAB).append("queue={").append(newline);
            for (LockRequest req : queue) {
                sb.append(TAB).append(TAB).append("LockRequest(").append(
                        newline);
                sb.append(TAB).append(TAB).append(TAB).append("status=")
                        .append(req.status).append(" mode=").append(req.mode)
                        .append(newline);
                sb.append(TAB).append(TAB).append(TAB).append("convertMode=")
                        .append(req.convertMode).append(" convertDuration=")
                        .append(req.convertDuration).append(newline);
                sb.append(TAB).append(TAB).append(TAB).append("owner=");
                if (req.owner instanceof Dumpable) {
                    ((Dumpable) req.owner).appendTo(sb);
                } else {
                    sb.append(req.owner);
                }
                sb.append(newline);
                sb.append(TAB).append(TAB).append(TAB).append("thread=")
                        .append(req.thread).append(newline);
                sb.append(TAB).append(TAB).append(")").append(newline);
            }
            sb.append(TAB).append("}").append(newline);
            return sb;
        }
    }

    static enum LockRequestStatus {
        GRANTED, CONVERTING, WAITING, DENIED;
    }

    /**
     * A LockRequest represents the request by a transaction for a lock.
     * 
     * @author Dibyendu Majumdar
     * 
     */
    static final class LockRequest extends Linkable {

        volatile LockRequestStatus status = LockRequestStatus.GRANTED;

        LockMode mode;

        LockMode convertMode;

        LockDuration convertDuration;

        volatile short count = 1;

        LockDuration duration;

        final Object owner;

        final LockItem lockItem;

        volatile Thread thread;

        LockRequest(LockItem lockItem, Object owner, LockMode mode,
                LockDuration duration) {
            this.lockItem = lockItem;
            this.mode = mode;
            this.convertMode = mode;
            this.duration = duration;
            this.thread = Thread.currentThread();
            this.owner = owner;
        }

        Object getOwner() {
            return owner;
        }

        @Override
        public String toString() {
            return "LockRequest(mode=" + mode + ", convertMode=" + convertMode
                    + ", status=" + status + ", count=" + count + ", owner="
                    + owner + ", thread=" + thread + ", duration=" + duration
                    + ")";
        }
    }

    /**
     * LockHandleImpl is an implementation of LockHandle interface. Since
     * LockHandles are stored in Transactions, these handles need to be as
     * compactly represented as possible.
     * 
     * @author Dibyendu Majumdar
     * @since Nov 11, 2005
     */
    static final class LockHandleImpl implements LockHandle {

        private final LockManagerImpl lockMgr;

        LockRequest lockRequest;

        LockHandleImpl(LockManagerImpl lockMgr, LockParameters parms) {
            this.lockMgr = lockMgr;
        }

        public final boolean release(boolean force) {
            return lockMgr.doRelease(this,
                    force ? LockManagerImpl.ReleaseAction.FORCE_RELEASE
                            : LockManagerImpl.ReleaseAction.RELEASE, null);
        }

        public final void downgrade(LockMode mode) {
            lockMgr.doRelease(this, LockManagerImpl.ReleaseAction.DOWNGRADE,
                    mode);
        }

        public final LockMode getCurrentMode() {
            if (lockRequest == null) {
                lockMgr.exceptionHandler.errorThrow(this.getClass().getName(),
                        "getCurrentMode", new LockException(
                                new MessageInstance(m_EC0006)));
            }
            return lockRequest.mode;
        }

        @Override
        public String toString() {
            return "LockHandleImpl(owner=" + getOwner() + ", target="
                    + getLockable() + ", currentMode=" + getCurrentMode() + ")";
        }

        Object getLockable() {
            return lockRequest.lockItem.target;
        }

        Object getOwner() {
            return lockRequest.owner;
        }
    }

    /**
     * Describe the status of a lock acquisition request.
     * 
     * @author Dibyendu Majumdar
     */
    public enum LockStatus {
        GRANTED, GRANTABLE, TIMEOUT, DEADLOCK
    }

    /**
     * Holds information regarding a lock wait. Purpose is to enable deadlock
     * detection.
     * 
     * @since 15 Nov 2006
     */
    static final class LockWaiter {
        final LockRequest myLockRequest;
        LockWaiter cycle;
        boolean visited = false;
        final Thread thread;

        LockWaiter(LockRequest r, Thread t) {
            myLockRequest = r;
            thread = t;
        }
    }

    static final class DeadlockDetector implements Runnable {

        final LockManagerImpl lockManager;

        public DeadlockDetector(LockManagerImpl lockManager) {
            this.lockManager = lockManager;
        }

        public void run() {
            if (lockManager.stop) {
                return;
            }
            lockManager.detectDeadlocks();
        }

    }

    public void setDeadlockDetectorInterval(int seconds) {
        this.deadlockDetectorInterval = seconds;
    }

}
